package edu.washington.cs.sounddetector.server.integration;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.sound.sampled.UnsupportedAudioFileException;

import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.impl.client.DefaultHttpClient;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import edu.washington.cs.sounddetector.server.api.SoundDetectorAPI;
import edu.washington.cs.sounddetector.server.api.exception.CantBuildClassifierException;
import edu.washington.cs.sounddetector.server.api.exception.DetectorDoesNotExistException;
import edu.washington.cs.sounddetector.server.api.exception.PhoneDoesNotExistException;
import edu.washington.cs.sounddetector.server.api.exception.SoundClassDoesNotExistException;
import edu.washington.cs.sounddetector.server.api.exception.SoundDetectorInternalErrorException;
import edu.washington.cs.sounddetector.server.api.exception.SoundDoesNotExistException;
import edu.washington.cs.sounddetector.server.api.object.LabeledSound;
import edu.washington.cs.sounddetector.server.api.object.Notification;
import edu.washington.cs.sounddetector.server.api.object.PeakEvent;
import edu.washington.cs.sounddetector.server.api.object.SoundClass;
import edu.washington.cs.sounddetector.server.api.object.SoundFormat;
import edu.washington.cs.sounddetector.server.testutil.Data;
import edu.washington.cs.sounddetector.server.testutil.Data.Doorbell;
import edu.washington.cs.sounddetector.server.testutil.Data.Unknown;
import edu.washington.cs.sounddetector.server.testutil.Helper;

public class SoundDetectorAPITest
{

    private SoundDetectorAPI api;
    private String detectorId;
    private String phoneId;
    private String registrationId;

    @Before
    public void setUp() throws Exception
    {
        HttpClient client = new DefaultHttpClient();
        URI serverURI = new URI(
        "http://localhost:8080/sound-detector-webapp");
        //"http://kiska.cs.washington.edu:8080/sound-detector");
        api = new SoundDetectorAPI(serverURI, client);
        detectorId = "Kitchen" + Math.random();
        phoneId = "12345" + Math.random();
        registrationId = null;
    }

    @After
    public void tearDown() throws ClientProtocolException, IOException,
            SoundDetectorInternalErrorException, DetectorDoesNotExistException,
            CantBuildClassifierException
    {
        try
        {
            api.getPendingNotifications(phoneId);
            api.unregisterReceiverPhone(phoneId);
            List<Long> ids = new ArrayList<Long>();
            for (List<Long> idList : api.getSounds(detectorId).values())
                ids.addAll(idList);
            api.untrain(ids);
        }
        catch (DetectorDoesNotExistException e)
        {
            // do nothing 
        }
        catch (PhoneDoesNotExistException e)
        {
            // do nothing 
        }
    }

    @Test
    public void testRegisterAndUnregisterReceiverPhone() throws ClientProtocolException,
            IOException, SoundDetectorInternalErrorException
    {
        api.registerReceiverPhone(phoneId, registrationId);
        api.unregisterReceiverPhone(phoneId);
    }

    @Test
    public void testTrainAndUntrain() throws UnsupportedAudioFileException, IOException,
            CantBuildClassifierException, SoundDetectorInternalErrorException
    {
        SoundFormat format = Helper.getSoundFormat(Data.Doorbell.D1.getPath());
        List<LabeledSound> trainingData = Helper.getTrainingSounds();

        List<Long> ids = api.train(detectorId, trainingData, format);
        assertEquals(trainingData.size(), ids.size());

        api.untrain(ids);
    }

    @Test(expected = PhoneDoesNotExistException.class)
    public void testGetPendingNotificationsNotRegistered() throws ClientProtocolException,
            PhoneDoesNotExistException, IOException, SoundDetectorInternalErrorException
    {
        api.getPendingNotifications(phoneId);
    }

    @Test
    public void testGetPendingNotificationsEmpty() throws ClientProtocolException, IOException,
            SoundDetectorInternalErrorException, PhoneDoesNotExistException
    {
        api.registerReceiverPhone(phoneId, registrationId);
        List<Notification> notifications = api.getPendingNotifications(phoneId);
        assert (notifications.isEmpty());
    }

    @Test
    public void testGetSounds() throws IOException, UnsupportedAudioFileException,
            CantBuildClassifierException, SoundDetectorInternalErrorException,
            DetectorDoesNotExistException
    {
        List<LabeledSound> data = Helper.getSmallSetOfSounds();
        SoundFormat format = Helper.getSoundFormat(Data.Doorbell.D1.getPath());
        List<Long> expectedIds = api.train(detectorId, data, format);
        Map<SoundClass, List<Long>> sounds = api.getSounds(detectorId);
        List<Long> actualIds = new ArrayList<Long>();
        for (List<Long> idList : sounds.values())
            actualIds.addAll(idList);
        Iterator<Long> iter = expectedIds.iterator();
        while (iter.hasNext())
        {
            Long id = iter.next();
            assertTrue(actualIds.contains(id));
            iter.remove();
        }
        assertTrue(expectedIds.isEmpty());
    }

    @Test
    public void testGetSound() throws IOException, UnsupportedAudioFileException,
            CantBuildClassifierException, SoundDetectorInternalErrorException,
            SoundDoesNotExistException
    {
        List<LabeledSound> data = Helper.getSmallSetOfSounds();
        SoundFormat format = Helper.getSoundFormat(Data.Doorbell.D1.getPath());
        List<Long> ids = api.train(detectorId, data, format);
        Long id = ids.get(0);
        byte[] sound = api.getSound(id);
        assertTrue(Arrays.equals(data.get(0).getSound(), sound));
    }

    @Test
    public void testGetOldNotificationsEmpty() throws ClientProtocolException, IOException,
            SoundDetectorInternalErrorException, PhoneDoesNotExistException
    {
        api.registerReceiverPhone(phoneId, registrationId);
        Date start = new Date(System.currentTimeMillis() - 10000);
        Date end = new Date(System.currentTimeMillis() + 10000);
        List<Notification> notifications = api.getOldNotifications(phoneId, start, end);
        assertTrue(notifications.isEmpty());
    }

    @Test
    public void testGetOldNotifications() throws ClientProtocolException, IOException,
            SoundDetectorInternalErrorException, UnsupportedAudioFileException,
            CantBuildClassifierException, PhoneDoesNotExistException
    {
        Date start = new Date(System.currentTimeMillis() - 10000);

        api.registerReceiverPhone(phoneId, registrationId);

        List<LabeledSound> data = Helper.getSmallSetOfSounds();
        SoundFormat format = Helper.getSoundFormat(Data.Doorbell.D1.getPath());
        api.train(detectorId, data, format);

        List<PeakEvent> peakEvents = Arrays.asList(new PeakEvent[] { new PeakEvent(0, 1, 2),
                new PeakEvent(1, 0, 3), new PeakEvent(2, 1, 0.5) });
        String predictedClass = api
                .classify(detectorId, data.get(0).getSound(), format, peakEvents);

        Date end = new Date(System.currentTimeMillis() + 10000);
        List<Notification> notifications = api.getOldNotifications(phoneId, start, end);
        assertTrue(notifications.isEmpty());
        api.getPendingNotifications(phoneId);

        end = new Date(System.currentTimeMillis() + 10000);
        notifications = api.getOldNotifications(phoneId, start, end);
        assertEquals(1, notifications.size());
        Notification notification = notifications.get(0);
        assertEquals(predictedClass, notification.getSoundClass().getSoundClass());
        assertEquals(detectorId, notification.getDetectorId());
        assertEquals(peakEvents, notification.getPeakEvents());
    }

    @Test
    public void testSetSoundClassProperties() throws ClientProtocolException,
            CantBuildClassifierException, IOException, SoundDetectorInternalErrorException,
            UnsupportedAudioFileException, DetectorDoesNotExistException,
            PhoneDoesNotExistException, SoundClassDoesNotExistException
    {
        api.registerReceiverPhone(phoneId, registrationId);

        List<LabeledSound> data = Helper.getSmallSetOfSounds();
        SoundFormat format = Helper.getSoundFormat(Data.Doorbell.D1.getPath());
        api.train(detectorId, data, format);

        List<SoundClass> soundClasses = new ArrayList<SoundClass>();
        soundClasses.add(new SoundClass(Doorbell.soundClass, 5));
        soundClasses.add(new SoundClass(Unknown.soundClass, 15));

        api.setSoundClassProperties(detectorId, soundClasses);

        List<LabeledSound> testSounds = Helper.getSmallSetOfSounds();
        List<PeakEvent> peakEvents = Arrays.asList(new PeakEvent[] { new PeakEvent(0, 1, 2),
                new PeakEvent(1, 0, 3), new PeakEvent(2, 1, 0.5) });
        for (LabeledSound testSound : testSounds)
        {
            api.classify(detectorId, testSound.getSound(), format, peakEvents);
        }

        List<Notification> notifications = api.getPendingNotifications(phoneId);
        for (Notification notification : notifications)
        {
            SoundClass actualClass = notification.getSoundClass();
            int index = soundClasses.indexOf(actualClass);
            SoundClass expectedClass = soundClasses.get(index);
            assertEquals(expectedClass, actualClass);
        }
    }

    @Test
    public void testRenameDetector() throws IOException, UnsupportedAudioFileException,
            CantBuildClassifierException, SoundDetectorInternalErrorException,
            DetectorDoesNotExistException
    {
        List<LabeledSound> data = Helper.getSmallSetOfSounds();
        SoundFormat format = Helper.getSoundFormat(Data.Doorbell.D1.getPath());
        List<Long> expectedIds = api.train(detectorId, data, format);

        String newDetectorId = "new" + detectorId;
        api.renameDetector(detectorId, newDetectorId);

        Map<SoundClass, List<Long>> sounds = api.getSounds(newDetectorId);

        List<Long> actualIds = new ArrayList<Long>();
        for (List<Long> idList : sounds.values())
            actualIds.addAll(idList);

        Iterator<Long> iter = expectedIds.iterator();
        while (iter.hasNext())
        {
            Long id = iter.next();
            assertTrue(actualIds.contains(id));
            iter.remove();
        }
        assertTrue(expectedIds.isEmpty());
    }

    @Test
    public void testSimpleDataflow() throws UnsupportedAudioFileException, IOException,
            CantBuildClassifierException, SoundDetectorInternalErrorException,
            PhoneDoesNotExistException
    {
        api.registerReceiverPhone(phoneId, registrationId);
        SoundFormat format = Helper.getSoundFormat(Data.Doorbell.D1.getPath());
        List<LabeledSound> trainingData = Helper.getTrainingSounds();

        List<Long> ids = api.train(detectorId, trainingData, format);
        assertEquals(trainingData.size(), ids.size());

        List<LabeledSound> testSounds = Helper.getTestSounds();
        int incorrect = 0;
        List<PeakEvent> peakEvents = Arrays.asList(new PeakEvent[] { new PeakEvent(0, 1, 2),
                new PeakEvent(1, 0, 3), new PeakEvent(2, 1, 0.5) });
        List<String> predictedClasses = new ArrayList<String>();
        for (LabeledSound testSound : testSounds)
        {
            String soundClass = api.classify(detectorId, testSound.getSound(), format, peakEvents);

            if (!soundClass.equals(testSound.getSoundClass()))
                incorrect++;

            System.out.printf("Predicted %s for %s\n", soundClass, testSound.getSoundClass());
            predictedClasses.add(soundClass);
        }
        /*
         * NOTE: this assumes - The Server is extracting the same features and
         * using same ML algorithm - The ML algorithm is deterministic
         * 
         * But it's a good enough test for now.
         */

        int total = testSounds.size();
        int correct = total - incorrect;
        assertEquals(2, incorrect);
        assertEquals(62, correct);

        // do some tests on getPendingNotifications
        List<Notification> notifications = api.getPendingNotifications(phoneId);
        assertEquals(testSounds.size(), notifications.size());
        assertTrue(api.getPendingNotifications(phoneId).isEmpty());
        for (int i = 0; i < predictedClasses.size(); i++)
        {
            String expectedClass = predictedClasses.get(i);
            Notification notification = notifications.get(i);
            assertEquals(expectedClass, notification.getSoundClass().getSoundClass());
            assertEquals(detectorId, notification.getDetectorId());
            assertEquals(peakEvents, notification.getPeakEvents());
        }
    }

}
