package edu.washington.cs.sounddetector.server.sound;

import static org.junit.Assert.assertEquals;
import jAudioFeatureExtractor.Aggregators.Aggregator;
import jAudioFeatureExtractor.AudioFeatures.FeatureExtractor;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Properties;

import org.junit.Before;
import org.junit.Test;

import edu.washington.cs.sounddetector.server.sound.SoundProcessorBuilder.Props;

public class SoundProcessorBuilderTest
{

    private SoundProcessorBuilder builder;
    private FeatureExtractor powerSpectrum;
    private FeatureExtractor spectralCentroid;
    private Aggregator mean;
    private Aggregator standardDeviation;

    @Before
    public void setUp()
    {
        this.builder = new SoundProcessorBuilder();
        this.powerSpectrum = new jAudioFeatureExtractor.AudioFeatures.PowerSpectrum();
        this.spectralCentroid = new jAudioFeatureExtractor.AudioFeatures.SpectralCentroid();
        this.mean = new jAudioFeatureExtractor.Aggregators.Mean();
        this.standardDeviation = new jAudioFeatureExtractor.Aggregators.StandardDeviation();
    }

    @Test
    public void testSetFeaturesArray()
    {
        List<FeatureExtractor> expected = Arrays.asList(powerSpectrum, spectralCentroid);
        builder.setFeatures(powerSpectrum, spectralCentroid);
        List<FeatureExtractor> actual = builder.getFeatures();
        assertFeaturesEqual(expected, actual);
    }

    @Test
    public void testSetAggregatorsArray()
    {
        List<Aggregator> expected = Arrays.asList(mean, standardDeviation);
        builder.setAggregators(mean, standardDeviation);
        List<Aggregator> actual = builder.getAggregators();
        assertAggregatorsEqual(expected, actual);
    }

    @Test
    public void testSetProperties()
    {
        Properties props = new Properties();
        props.setProperty(Props.WINDOW_SIZE, "9");
        props.setProperty(Props.WINDOW_OVERLAP, "1");
        props.setProperty(Props.SAMPLING_RATE, "16000");
        props.setProperty(Props.NORMALIZE, "true");
        props.setProperty(Props.FEATURES, "Magnitude Spectrum");
        props.setProperty(Props.AGGREGATORS, "Area Moments");
        builder.setProperties(props);

        assertEquals(9, builder.getWindowSize());
        assertEquals(1, builder.getWindowOverlap(), 0.01);
        assertEquals(16000, builder.getSamplingRate(), 0.01);
        assertEquals(true, builder.isNormalize());
        assertEquals(1, builder.getFeatures().size());
        assertEquals("Magnitude Spectrum", builder.getFeatures().get(0).getFeatureDefinition().name);
        assertEquals(1, builder.getAggregators().size());
        assertEquals("Area Moments", builder.getAggregators().get(0).getAggregatorDefinition().name);
    }

    private void assertFeaturesEqual(List<FeatureExtractor> expected, List<FeatureExtractor> actual)
    {
        List<String> expectedNames = convertToSortedFeatureNames(expected);
        List<String> actualNames = convertToSortedFeatureNames(actual);
        assertEquals(expectedNames, actualNames);
    }

    private void assertAggregatorsEqual(List<Aggregator> expected, List<Aggregator> actual)
    {
        List<String> expectedNames = convertToSortedAggregatorNames(expected);
        List<String> actualNames = convertToSortedAggregatorNames(actual);
        assertEquals(expectedNames, actualNames);
    }

    private List<String> convertToSortedFeatureNames(List<FeatureExtractor> features)
    {
        List<String> names = new ArrayList<String>();
        for (FeatureExtractor feature : features)
        {
            names.add(feature.getFeatureDefinition().name);
        }
        Collections.sort(names);
        return names;
    }

    private List<String> convertToSortedAggregatorNames(List<Aggregator> aggregators)
    {
        List<String> names = new ArrayList<String>();
        for (Aggregator aggregator : aggregators)
        {
            names.add(aggregator.getAggregatorDefinition().name);
        }
        Collections.sort(names);
        return names;
    }

}
